/******************************************************************************

    $Id: buffer.c,v 1.2 2003/08/28 07:40:24 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/buffer.c,v $

******************************************************************************/

#include "priv.h"


void cpp_buffer_init(CPP* cpp, BUFFER* buf)
{
    size_t n = 256;
    memset(buf, 0, sizeof(BUFFER));
    buf->cpp = cpp;
    buf->data = (char*)cpp_safe_malloc(cpp, n);
    buf->size = n;
}


void cpp_buffer_cleanup(BUFFER* buf)
{
    cpp_free(buf->cpp, buf->data);
    buf->data = NULL;
    buf->size = 0;
    buf->len = 0;
}


void cpp_buffer_grow(BUFFER* buf, size_t need)
{
    if ( need > buf->size ) {
        size_t n = buf->size;
        while ( n < need ) {
            if ( (n<<=1) == 0 ) {
                n = need;
            }
        }
        buf->data = (char*)cpp_safe_realloc(buf->cpp, buf->data, n);
        buf->size = n;
    }
}


void cpp_buffer_clear(BUFFER* buf)
{
    buf->len = 0;
}


void cpp_buffer_catc(BUFFER* buf, char ch)
{
    if ( buf->len >= buf->size ) {
        cpp_buffer_grow(buf, buf->len+1);
    }
    buf->data[buf->len++] = ch;
}


void cpp_buffer_catn(BUFFER* buf, const char* str, size_t len)
{
    size_t need = buf->len + len;
    if ( need >= buf->size ) {
        cpp_buffer_grow(buf, need);
    }
    memcpy(buf->data + buf->len, str, len);
    buf->len += len;
}


void cpp_buffer_cats(BUFFER* buf, const char* str)
{
    cpp_buffer_catn(buf, str, strlen(str));
}


void cpp_buffer_cati(BUFFER* buf, int number)
{
    char buffer[sizeof(number)*3+2];
    sprintf(buffer, "%d", number);
    cpp_buffer_cats(buf, buffer);
}


void cpp_buffer_insertc(BUFFER* buf, size_t pos, char ch)
{
    size_t need = buf->len + 1;
    assert(pos <= buf->len);
    if ( need >= buf->size ) {
        cpp_buffer_grow(buf, need);
    }

    memmove(buf->data + pos, buf->data + pos + 1, buf->len - pos);
    buf->data[pos] = ch;
    buf->len++;
}


void cpp_buffer_removec(BUFFER* buf, size_t pos)
{
    assert(pos < buf->len);

    memmove(buf->data + pos, buf->data + pos + 1, buf->len - pos - 1);
    buf->len--;
}


char* cpp_buffer_gets(BUFFER* buf)
{
    if ( buf->len >= buf->size ) {
        cpp_buffer_grow(buf, buf->len+1);
    }
    buf->data[buf->len] = '\0';
    return buf->data;
}
