/******************************************************************************

    $Id: config.c,v 1.2 2003/08/28 19:12:27 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/config.c,v $

******************************************************************************/

#include "priv.h"


void cpp_config_define(CPP* cpp, const char* name, size_t len, const char* value)
{
    size_t      i;
    DEFINE**    defp;
    DEFINE*     def;

    // Validate identifier
    i = 0;
    if ( len ) {
        if ( ISIDENT1(name[0]) ) {
            ++i;
            while ( i < len && ISIDENT(name[i]) ) {
                ++i;
            }
        }
    }
    if ( i == 0 || i != len ) {
        cpp_fatal(cpp, "Identifier expected");
    }

    // Create our own copy of the value
    if ( value ) {
        value = cpp_heap_strdup(&cpp->cpp_heap, value);
    }

    // If name is already listed, replace value
    for ( defp = &cpp->defines ; (def=*defp) != NULL ; defp = &def->next ) {
        if ( len == strlen(def->name) && memcmp(name, def->name, len) == 0 ) {
            def->value = value;
            return;
        }
    }

    // It wasn't there, append to list
    def = (DEFINE*)cpp_heap_calloc(&cpp->cpp_heap, 1, sizeof(DEFINE));
    def->name = cpp_heap_strndup(&cpp->cpp_heap, name, len);
    def->value = value;
    *defp = def;
}


void cpp_config_include(CPP* cpp, const char* path)
{
    INCDIR**    dirp;
    INCDIR*     dir;
    char*       clone;
    size_t      len;
    
    // Validate path
    len = strlen(path);
    if ( len == 0 ) {
        cpp_fatal(cpp, "Expected directory path");
    }

    // Create our own copy with a trailing directory separator
    clone = cpp_heap_malloc(&cpp->cpp_heap, len+2);
    strcpy(clone, path);
    if ( !cpp_is_dirsep(clone[len-1]) ) {
        clone[len++] = cpp_dirsep();
        clone[len] = '\0';
    }

    // See if directory is already listed - if so, ignore it
    for ( dirp = &cpp->includes ; (dir=*dirp) != NULL ; dirp = &dir->next ) {
        if ( cpp_pathcmp(dir->path, clone) == 0 ) {
            return;
        }
    }

    // Append directory to list
    dir = (INCDIR*)cpp_heap_calloc(&cpp->cpp_heap, 1, sizeof(INCDIR));
    dir->path = clone;
    *dirp = dir;
}


void cpp_config_include_list(CPP* cpp, const char* list)
{
    char        listsep = cpp_listsep();
    const char* cp;
        
    if ( list ) {
        for (;;) {
            while ( *list == listsep ) {
                ++list;
            }
            if ( *list == '\0' ) {
                break;
            }
            cp = strchr(list, listsep);
            if ( cp == NULL ) {
                cpp_config_include(cpp, list);
                break;
            }
            cpp_buffer_clear(&cpp->scratch);
            cpp_buffer_catn(&cpp->scratch, list, cp-list);
            cpp_config_include(cpp, cpp_buffer_gets(&cpp->scratch));
            list = cp+1;
        }
    }
}
