/******************************************************************************

    $Id: cpp.h,v 1.5 2003/09/05 20:30:27 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/cpp.h,v $

    Public interface to preprocessor library.

******************************************************************************/

// Includes
#include <stddef.h>
#include <stdio.h>

// Tokens
enum {
    // Special tokens
    T_END,          //  End of input
    T_SPACE,        //  White space
    T_PRAGMA,       //  'str' contains pragma string
    T_ERROR,        //  'str' contains error message
    // Regular source file tokens
    T_IDENT,        //  Identifier
    T_NUMLIT,       //  Numeric literal
    T_STRLIT,       //  "string literal"
    T_CHRLIT,       //  'character literal'
    T_NOT,          //  !
    T_NE,           //  !=
    T_HASH,         //  #
    T_HASH_HASH,    //  ##
    T_MOD,          //  %
    T_MOD_ASSIGN,   //  %=
    T_AND,          //  &
    T_AND_AND,      //  &&
    T_AND_ASSIGN,   //  &=
    T_PAREN_OPEN,   //  (
    T_PAREN_CLOSE,  //  )
    T_MUL,          //  *
    T_MUL_ASSIGN,   //  *=
    T_ADD,          //  +
    T_INC,          //  ++
    T_ADD_ASSIGN,   //  +=
    T_COMMA,        //  ,
    T_SUB,          //  -
    T_DEC,          //  --
    T_SUB_ASSIGN,   //  -=
    T_ARROW,        //  ->
    T_DOT,          //  .
    T_ELLIPSIS,     //  ...
    T_DIV,          //  /
    T_DIV_ASSIGN,   //  /=
    T_COLON,        //  :
    T_SEMI,         //  ;
    T_LT,           //  <
    T_SHL,          //  <<
    T_SHL_ASSIGN,   //  <<=
    T_LE,           //  <=
    T_ASSIGN,       //  =
    T_EQ,           //  ==
    T_GT,           //  >
    T_GE,           //  >=
    T_SHR,          //  >>
    T_SHR_ASSIGN,   //  >>=
    T_QMARK,        //  ?
    T_SHARP_OPEN,   //  [
    T_SHARP_CLOSE,  //  ]
    T_XOR,          //  ^
    T_XOR_ASSIGN,   //  ^=
    T_CURLY_OPEN,   //  {
    T_OR,           //  |
    T_OR_ASSIGN,    //  |=
    T_OR_OR,        //  ||
    T_CURLY_CLOSE,  //  }
    T_NEG,          //  ~
    T_BAD           //  Bad character
};

// Forwards
typedef struct tagTOKEN TOKEN;
typedef struct tagCPP   CPP;

// Scanned token
struct tagTOKEN {
    const char*     file;   // __FILE__ where found
    int             line;   // __LINE__ where found
    int             tag;    // Tag of token
    const char*     str;    // Text from source file
};

// Instantiate and release
CPP* cpp_new(void);
void cpp_delete(CPP* cpp);

// Configure preprocessor
const char* cpp_define(CPP* cpp, const char* assignment);
const char* cpp_undefine(CPP* cpp, const char* name);
const char* cpp_include(CPP* cpp, const char* path);
const char* cpp_include_list(CPP* cpp, const char* list);

// Preprocessing
const char* cpp_open(CPP* cpp, const char* path, FILE* fp);
const char* cpp_gettoken(CPP* cpp, TOKEN* token);

// Utility functions
const char* cpp_strtag(int tag);
void cpp_dump_macros(CPP* cpp, FILE* fp, int verbose);
void cpp_dump_memstat(CPP* cpp, FILE* fp);
