/******************************************************************************

    $Id: dump.c,v 1.3 2003/08/28 17:43:44 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/dump.c,v $

******************************************************************************/

#include "priv.h"
#include "expr.h"


void cpp_dump_expression(int level, EXPR* expr)
{
    FILE* fp = stdout;

    switch ( expr->type ) {
        case E_VALUE:
            if ( expr->u.value.is_signed ) {
                fprintf(fp, "%*sVALUE %ld\n", level*2, "", expr->u.value.u.svalue);
            } else {
                fprintf(fp, "%*sVALUE %luu\n", level*2, "", expr->u.value.u.uvalue);
            }
            break;
        case E_CONDITIONAL:
            fprintf(fp, "%*sCONDITIONAL\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.conditional.condition);
            cpp_dump_expression(level+1, expr->u.conditional.iftrue);
            cpp_dump_expression(level+1, expr->u.conditional.iffalse);
            break;
        case E_LOGIC_OR:
            fprintf(fp, "%*sLOGIC_OR\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_LOGIC_AND:
            fprintf(fp, "%*sLOGIC_AND\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_BIT_OR:
            fprintf(fp, "%*sBIT_OR\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_BIT_XOR:
            fprintf(fp, "%*sBIT_XOR\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_BIT_AND:
            fprintf(fp, "%*sBIT_AND\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_SHL:
            fprintf(fp, "%*sSHL\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_SHR:
            fprintf(fp, "%*sSHR\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_EQ:
            fprintf(fp, "%*sEQ\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_NE:
            fprintf(fp, "%*sNE\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_GT:
            fprintf(fp, "%*sGT\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_GE:
            fprintf(fp, "%*sGE\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_LT:
            fprintf(fp, "%*sLT\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_LE:
            fprintf(fp, "%*sLE\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_ADD:
            fprintf(fp, "%*sADD\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_SUB:
            fprintf(fp, "%*sSUB\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_MUL:
            fprintf(fp, "%*sMUL\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_DIV:
            fprintf(fp, "%*sDIV\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_MOD:
            fprintf(fp, "%*sMOD\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.binary.left);
            cpp_dump_expression(level+1, expr->u.binary.right);
            break;
        case E_UPLUS:
            fprintf(fp, "%*sUPLUS\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.unary.operand);
            break;
        case E_UMINUS:
            fprintf(fp, "%*sUMINUS\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.unary.operand);
            break;
        case E_NOT:
            fprintf(fp, "%*sNOT\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.unary.operand);
            break;
        case E_NEG:
            fprintf(fp, "%*sNEG\n", level*2, "");
            cpp_dump_expression(level+1, expr->u.unary.operand);
            break;
        default:
            fprintf(fp, "%*sUnknown %d\n", level*2, "", expr->type);
            break;
    }
}


void cpp_dump_macro(MACRO* macro, FILE* fp, int verbose)
{
    size_t      i;
    MACROTOKEN* token;
    char        stats[64];

    switch ( macro->expand_count ) {
        case 0:
            strcpy(stats, "unused");
            break;
        case 1:
            strcpy(stats, "1 expansion");
            break;
        default:
            sprintf(stats, "%d expansions", macro->expand_count);
            break;
    }

    fprintf(fp, "#line %d \"%s\" /* %s */ \n", macro->line, macro->file, stats);
    fprintf(fp, "#define %s", macro->name);
    if ( macro->fnlike ) {
        fprintf(fp, "(");
        for ( i = 0 ; i < macro->argc ; i++ ) {
            if ( i ) {
                fprintf(fp, ",");
            }
            fprintf(fp, macro->argv[i]);
        }
        fprintf(fp, ")");
    }
    fprintf(fp, " ");

    if ( verbose ) {
        char comment[20];
        fprintf(fp, "\\\n");
        for ( token = macro->body ; token ; token = token->next ) {
            if ( token->tag != T_SPACE ) {
                switch ( token->tag ) {
                    case T_STRINGIZE:
                        sprintf(comment, "/* #arg %u */", token->argno);
                        break;
                    case T_ARGUMENT:
                        sprintf(comment, "/* arg %u */", token->argno);
                        break;
                    default:
                        comment[0] = '\0';
                        break;
                }
                fprintf(fp, "    %-18s%s", comment, token->str);
                if ( token->next ) {
                    fprintf(fp, " \\");
                }
                fprintf(fp, "\n");
            }
        }
    } else {
        for ( token = macro->body ; token ; token = token->next ) {
            if ( token->tag == T_HASH_HASH ) {
                fprintf(fp, " %s ", token->str);
            } else {
                fprintf(fp, "%s", token->str);
            }
        }
    }

    fprintf(fp, "\n\n");
}


int cpp_macro_cmp(const void* v1, const void* v2)
{
    const MACRO** m1 = (const MACRO**)v1;
    const MACRO** m2 = (const MACRO**)v2;

    return strcmp((*m1)->name, (*m2)->name);
}


void cpp_dump_macros(CPP* cpp, FILE* fp, int verbose)
{
    MACRO** vector;
    size_t  i;
    size_t  j;
    MACRO*  macro;

    if ( cpp->macro_count ) {
        vector = cpp_malloc(cpp, cpp->macro_count * sizeof(MACRO*));
        if ( vector ) {
            for ( i = j = 0 ; i < cpp->macro_hsize ; i++ ) {
                for ( macro = cpp->macro_htab[i] ; macro ; macro = macro->next ) {
                    vector[j++] = macro;
                }
            }
            assert(j == cpp->macro_count);
            qsort(vector, j, sizeof(MACRO*), cpp_macro_cmp);
            for ( j = 0 ; j < cpp->macro_count ; j++ ) {
                cpp_dump_macro(vector[j], fp, verbose);
            }
            cpp_free(cpp, vector);
        } else {
            for ( i = j = 0 ; i < cpp->macro_hsize ; i++ ) {
                for ( macro = cpp->macro_htab[i] ; macro ; macro = macro->next ) {
                    cpp_dump_macro(macro, fp, verbose);
                }
            }
        }
    }
}


void cpp_dump_memstat(CPP* cpp, FILE* fp)
{
    fprintf(fp, "Memory statistics\n");
    fprintf(fp, "    System memory allocations........ %8u malloc/calloc/realloc\n", (unsigned)cpp->alloc_count);
    fprintf(fp, "    System memory deallocations...... %8u free/realloc\n", (unsigned)cpp->dealloc_count);
    fprintf(fp, "    Input line buffer size........... %8u characters\n", (unsigned)cpp->input_buf.size);
    fprintf(fp, "    Scratch buffer size.............. %8u characters\n", (unsigned)cpp->scratch.size);
    fprintf(fp, "    Message buffer size.............. %8u characters\n", (unsigned)cpp->msg_buf.size);
    fprintf(fp, "    Macro hash table size............ %8u buckets\n", (unsigned)cpp->macro_hsize);
    fprintf(fp, "    Macro definitions................ %8u macros\n", (unsigned)cpp->macro_count);
    fprintf(fp, "    Stack size for #if-blocks........ %8u slots\n", (unsigned)cpp->cond_stack_size);
    fprintf(fp, "    Heap for preprocessor config..... %8u bytes (top %u)\n", (unsigned)cpp->cpp_heap.size, (unsigned)cpp->cpp_heap.top_size);
    fprintf(fp, "    Heap for translation unit........ %8u bytes (top %u)\n", (unsigned)cpp->unit_heap.size, (unsigned)cpp->unit_heap.top_size);
    fprintf(fp, "    Heap for temporary output........ %8u bytes (top %u)\n", (unsigned)cpp->output_heap.size, (unsigned)cpp->output_heap.top_size);
}
