/******************************************************************************

    $Id: expand.c,v 1.5 2003/08/28 21:59:29 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/expand.c,v $

******************************************************************************/

#include "priv.h"


static bool cpp_is_next_parenopen(PPTOKEN* list)
{
    list = cpp_skip_t_space(list->next);
    return list && list->token.tag == T_PAREN_OPEN;
}


static void cpp_too_many_arguments(CPP* cpp, MACRO* macro)
{
    cpp_msg_new(cpp);
    cpp_msg_cats(cpp, "Too many arguments for macro '");
    cpp_msg_cats(cpp, macro->name);
    cpp_msg_cats(cpp, "'");
    cpp_output_append(cpp, cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp)));
}


static void cpp_too_few_arguments(CPP* cpp, MACRO* macro)
{
    cpp_msg_new(cpp);
    cpp_msg_cats(cpp, "Too many arguments for macro '");
    cpp_msg_cats(cpp, macro->name);
    cpp_msg_cats(cpp, "'");
    cpp_output_append(cpp, cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp)));
}


static void cpp_open_arglist(CPP* cpp, MACRO* macro)
{
    cpp_msg_new(cpp);
    cpp_msg_cats(cpp, "Argument list to macro '");
    cpp_msg_cats(cpp, macro->name);
    cpp_msg_cats(cpp, "' not closed with ')'");
    cpp_output_append(cpp, cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp)));
}


static PPTOKEN* cpp_macro_paste(CPP* cpp, PPTOKEN** argv, size_t prev_argno, PPTOKEN* prev, MACROTOKEN* next)
{
    BUFFER*     buf;
    PPLIST      result;
    PPTOKEN*    token;
    const char* pos;
    int         tag;
    size_t      len;
    bool        use_marker = false;
    
    // Expand both operands into string representation
    buf = &cpp->scratch;
    cpp_buffer_clear(buf);
    if ( prev->token.tag == T_ARGUMENT ) {
        use_marker = true;
        cpp_list_flatten(buf, argv[prev_argno]);
    } else {
        cpp_buffer_cats(buf, prev->token.str);
    }
    if ( next->tag == T_ARGUMENT ) {
        use_marker = true;
        cpp_list_flatten(buf, argv[next->argno]);
    } else {
        cpp_buffer_cats(buf, next->str);
    }

    // Restan string
    pos = cpp_buffer_gets(buf);
    cpp_list_reset(&result);
    while ( cpp_identify_token(pos,&tag,&len), tag != T_END ) {
        token = cpp_token_newn(cpp, tag, pos, len);
        cpp_list_append(&result, token);
        pos += len;
    }

    // Create place marker if result is empty
    if ( result.first==NULL && use_marker ) {
        cpp_list_append(&result, cpp_token_new(cpp, T_PLACEMARKER));
    }

    return result.first;
}


static PPTOKEN* cpp_macro_stringize(CPP* cpp, PPTOKEN* list)
{
    BUFFER*     buf;
    const char* pos;

    buf = &cpp->scratch;
    cpp_buffer_clear(buf);
    cpp_buffer_catc(buf, '"');
    while ( list ) {
        pos = list->token.str;
        while ( *pos ) {
            if ( *pos == '"' || *pos == '\\' ) {
                cpp_buffer_catc(buf, '\\');
            }
            cpp_buffer_catc(buf, *pos++);
        }
        list = list->next;
    }
    cpp_buffer_catc(buf, '"');

    return cpp_token_news(cpp, T_STRLIT, cpp_buffer_gets(buf));
}


void cpp_list_flatten(BUFFER* dest, PPTOKEN* list)
{
    while ( list ) {
        cpp_buffer_cats(dest, list->token.str);
        list = list->next;
    }
}


PPTOKEN* cpp_expand_list(CPP* cpp, PPTOKEN* list, FRAME* parent)
{
    PPLIST      result;
    MACRO*      macro;
    PPTOKEN*    token;
    PPTOKEN**   argv;
    PPTOKEN*    expanded;
    size_t      argno;
    int         paren;

    cpp_list_reset(&result);

    while ( list ) {
        if ( list->token.tag != T_IDENT
          || (macro=cpp_find_macro(cpp,list->token.str)) == NULL
          || (macro->fnlike && !cpp_is_next_parenopen(list))
           )
        {
            // Clone token
            token = cpp_token_clone(cpp, list);
            cpp_list_append(&result, token);
            list = list->next;
        } else {
            // Expand macro
            argv = NULL;
            list = list->next;
            if ( macro->fnlike ) {
                // Parse argument list
                list = cpp_skip_t_space(list)->next;
                argv = cpp_heap_calloc(&cpp->output_heap, macro->argc, sizeof(PPTOKEN*));
                paren = 0;
                argno = 0;
                while ( list && (list->token.tag != T_PAREN_CLOSE || paren) ) {
                    switch ( list->token.tag ) {
                        case T_PAREN_OPEN:
                            ++paren;
                            break;
                        case T_PAREN_CLOSE:
                            --paren;
                            break;
                        case T_COMMA:
                            if ( paren == 0 ) {
                                if ( argno +1 < macro->argc || !macro->varargs ) {
                                    ++argno;
                                    list = list->next;
                                    continue;
                                }
                            }
                            break;
                    }
                    if ( list->token.tag != T_SPACE && macro->argc == 0 ) {
                        cpp_too_many_arguments(cpp, macro);
                        return NULL;
                    }
                    if ( argno < macro->argc ) {
                        token = cpp_token_clone(cpp, list);
                        token->next = argv[argno];
                        argv[argno] = token;
                    }
                    list = list->next;
                }
                if ( list == NULL ) {
                    cpp_open_arglist(cpp, macro);
                    return NULL;
                }
                assert(list->token.tag == T_PAREN_CLOSE);
                list = list->next;
                if ( macro->argc != 0 ) {
                    ++argno;
                }
                if ( argno < macro->argc ) {
                    cpp_too_few_arguments(cpp, macro);
                    return NULL;
                }
                if ( argno > macro->argc ) {
                    cpp_too_many_arguments(cpp, macro);
                    return NULL;
                }
                for ( argno = 0 ; argno < macro->argc ; argno++ ) {
                    argv[argno] = cpp_skip_t_space(argv[argno]);
                    argv[argno] = cpp_reverse_tokens(argv[argno]);
                    argv[argno] = cpp_skip_t_space(argv[argno]);
                }
            }
            expanded = cpp_expand_macro(cpp, macro, argv, parent);
            cpp_list_appendlist(&result, expanded);
        }
    }

    cpp_listfix(cpp, &result);

    return result.first;
}


PPTOKEN* cpp_expand_macro(CPP* cpp, MACRO* macro, PPTOKEN** argv, FRAME* parent)
{
    FRAME       frame;
    PPTOKEN*    rev_list;
    PPTOKEN*    list;
    PPTOKEN*    result;
    PPTOKEN*    token;
    MACROTOKEN* mtoken;
    size_t      prev_argno = (size_t)-1;

    // Initialize
    frame.parent = parent;
    frame.macro  = macro;
    frame.name   = macro->name;

    macro->expand_count++;

    // Expand replacement list - built in reverse order
    rev_list = NULL;
    if ( macro->special ) {
        rev_list = cpp_special_macro(cpp, macro);
    } else {
        for ( mtoken = macro->body ; mtoken ; mtoken = mtoken->next ) {
            switch ( mtoken->tag ) {
                case T_ARGUMENT:
                    if ( mtoken->next && mtoken->next->tag == T_HASH_HASH ) {
                        // argument is pasted when we reach next token
                        token = cpp_token_new(cpp, T_ARGUMENT);
                        token->next = rev_list;
                        rev_list = token;
                        prev_argno = mtoken->argno; // save for evaluation of '##'
                    } else {
                        list = cpp_expand_list(cpp, argv[mtoken->argno], NULL);
                        if ( list ) {
                            while ( list ) {
                                token = list;
                                list = list->next;
                                token->next = rev_list;
                                rev_list = token;
                            }
                        } else {
                            token = cpp_token_new(cpp, T_PLACEMARKER);
                            token->next = rev_list;
                            rev_list = token;
                        }
                    }
                    break;
                case T_STRINGIZE:
                    token = cpp_macro_stringize(cpp, argv[mtoken->argno]);
                    token->next = rev_list;
                    rev_list = token;
                    break;
                case T_HASH_HASH:
                    // Combine tokens and rescan them
                    mtoken = mtoken->next;
                    list = cpp_macro_paste(cpp, argv, prev_argno, rev_list, mtoken);
                    rev_list = rev_list->next; // get rid of previous token
                    while ( list ) {
                        token = list;
                        list = list->next;
                        token->next = rev_list;
                        rev_list = token;
                    }
                    break;
                default:
                    token = cpp_token_news_recycle(cpp, mtoken->tag, mtoken->str);
                    token->next = rev_list;
                    rev_list = token;
                    break;
            }
        }
    }

    // Process results:
    // - list is reversed
    // - identifiers are marked as final, if on expansion stack
    // - place markers are removed
    rev_list = cpp_skip_t_space(rev_list);
    result = NULL;
    while ( rev_list ) {
        token = rev_list;
        rev_list = rev_list->next;
        if ( token->token.tag != T_PLACEMARKER ) {
            token->next = result;
            result = token;
            if ( token->token.tag == T_IDENT ) {
                const char* ident = token->token.str;
                FRAME* f;
                for ( f = &frame ; f ; f = f->parent ) {
                    if ( strcmp(ident, f->name) == 0 ) {
                        token->token.tag = T_FINAL_IDENT;
                        break;
                    }
                }
            }
        }
    }
    result = cpp_skip_t_space(result);

    // Return rescanned macro
    return cpp_expand_list(cpp, result, &frame);
}
