/******************************************************************************

    $Id: include.c,v 1.1 2003/08/28 19:12:27 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/include.c,v $

******************************************************************************/

#include "priv.h"


void cpp_parse_include(CPP* cpp, const char* pos)
{
    PPTOKEN*    list;
    int         tag;
    size_t      len;
    char        quote;
    const char* name;
    const char* path;
    INCDIR*     incdir;

    pos = cpp_skipspace(pos);

    // Expand macros if the does not contain valid header name (only)
    len = cpp_header_name_len(pos);
    if ( len == 0 || *cpp_skipspace(pos+len) ) {
        list = cpp_tokenize(cpp, pos);
        list = cpp_expand_list(cpp, list, NULL);
        cpp_buffer_clear(&cpp->scratch);
        cpp_list_flatten(&cpp->scratch, list);
        pos = cpp_buffer_gets(&cpp->scratch);
        pos = cpp_skipspace(pos);
        len = cpp_header_name_len(pos);
        if ( len == 0 ) {
            cpp_identify_token(pos, &tag, &len);
            cpp_msg_new(cpp);
            cpp_msg_cats(cpp, "Expected header name after '#include'");
            if ( tag != T_END ) {
                cpp_msg_cats(cpp, ", found ");
                cpp_msg_cats(cpp, cpp_strtag(tag));
            }
            cpp_output_append(cpp, cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp)));
            return;
        }
        if ( *cpp_skipspace(pos+len) ) {
            pos = cpp_skipspace(pos+len);
            cpp_identify_token(pos, &tag, &len);
            cpp_msg_new(cpp);
            cpp_msg_cats(cpp, "Unexpected token after header name: ");
            cpp_msg_cats(cpp, cpp_strtag(tag));
            cpp_output_append(cpp, cpp_token_news(cpp, T_ERROR, cpp_msg_gets(cpp)));
            return;
        }
    }

    quote = *pos;
    name = cpp_heap_strndup(&cpp->output_heap, pos+1, len-2);

    if ( cpp_is_path_abs(name) ) {
        // No search - just use the name
        path = cpp_heap_strdup(&cpp->unit_heap, name);
    } else {
        path = NULL;
        // Try local directory if quoted name
        if ( quote == '"' ) {
            const char* current_path = cpp->input->path;
            const char* current_base = cpp_basename(current_path);
            cpp_buffer_clear(&cpp->scratch);
            cpp_buffer_catn(&cpp->scratch, current_path, current_base - current_path);
            cpp_buffer_cats(&cpp->scratch, name);
            if ( cpp_exists(cpp_buffer_gets(&cpp->scratch)) ) {
                // Got it!
                path = cpp_heap_strdup(&cpp->unit_heap, cpp_buffer_gets(&cpp->scratch));
            }
        }
        // Try include path if not found
        for ( incdir = cpp->includes ; path==NULL && incdir ; incdir = incdir->next ) {
            cpp_buffer_clear(&cpp->scratch);
            cpp_buffer_cats(&cpp->scratch, incdir->path);
            cpp_buffer_cats(&cpp->scratch, name);
            if ( cpp_exists(cpp_buffer_gets(&cpp->scratch)) ) {
                // Got it!
                path = cpp_heap_strdup(&cpp->unit_heap, cpp_buffer_gets(&cpp->scratch));
            }
        }
        if ( path == NULL ) {
            cpp_msg_newfl(cpp, cpp->file, cpp->line);
            cpp_msg_cats(cpp, "File not found ");
            cpp_msg_catn(cpp, pos, len);
            cpp_fatal(cpp, cpp_msg_gets(cpp));
        }
    }

    // Stack input file
    cpp_input_push(cpp, path);
    cpp->input->fp = fopen(path, "r");
    cpp->input->owner = true;
    if ( cpp->input->fp == NULL ) {
        cpp_msg_newfl(cpp, cpp->file, cpp->line);
        cpp_msg_cats(cpp, "Error opening \"");
        cpp_msg_cats(cpp, path);
        cpp_msg_cats(cpp, "\": ");
        cpp_msg_cats(cpp, strerror(errno));
        cpp_fatal(cpp, cpp_msg_gets(cpp));
    }
}
