/******************************************************************************

    $Id: line.c,v 1.1 2003/08/28 09:53:28 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/line.c,v $

******************************************************************************/

#include "priv.h"


typedef struct tagLOCATION LOCATION;

struct tagLOCATION {
    const char* line_pos;
    size_t      line_len;
    const char* file_pos;
    size_t      file_len;
};


const char* cpp_entok_line_directive(CPP* cpp, const char* pos, LOCATION* tokens)
{
    int     tag;
    size_t  len;

    pos = cpp_skipspace(pos);
    len = cpp_digit_sequence_len(pos);
    if ( len == 0 ) {
        return "Line number expected after '#line'";
    }
    tokens->line_pos = pos;
    tokens->line_len = len;
    pos = cpp_skipspace(pos+len);

    if ( *pos == '\0' ) {
        tokens->file_pos = NULL;
        tokens->file_len = 0;
        return NULL;
    }

    if ( *pos != '"' || (len=cpp_header_name_len(pos))==0 ) {
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "Quoted file name expected after '#line ");
        cpp_msg_catn(cpp, tokens->line_pos, tokens->line_len);
        return cpp_msg_gets(cpp);
    }

    tokens->file_pos = pos;
    tokens->file_len = len;

    pos = cpp_skipspace(pos+len);
    if ( *pos ) {
        cpp_identify_token(pos, &tag, &len);
        cpp_msg_new(cpp);
        cpp_msg_cats(cpp, "Unexpected token in '#line' directive: ");
        cpp_msg_cats(cpp, cpp_strtag(tag));
        return cpp_msg_gets(cpp);
    }

    return NULL;
}



void cpp_parse_line(CPP* cpp, const char* pos)
{
    PPTOKEN*    list;
    const char* msg;
    LOCATION    loc;
    int         line;
    const char* file;

    msg = cpp_entok_line_directive(cpp, pos, &loc);
    if ( msg ) {
        list = cpp_tokenize(cpp, pos);
        list = cpp_expand_list(cpp, list, NULL);
        cpp_buffer_clear(&cpp->scratch);
        cpp_list_flatten(&cpp->scratch, list);
        pos = cpp_buffer_gets(&cpp->scratch);
        msg = cpp_entok_line_directive(cpp, pos, &loc);
        if ( msg ) {
            cpp_output_append(cpp, cpp_token_news(cpp, T_ERROR, msg));
            return;
        }
    }

    // Make sure exactly 1 newline is emitted after this line
    cpp->newlines = cpp_token_newnl(cpp, 1);

    line = atoi(loc.line_pos);
    file = NULL;
    if ( loc.file_pos && loc.file_len > 2 ) {
        file = cpp_heap_strndup(&cpp->unit_heap, loc.file_pos+1, loc.file_len-2);
    }

    cpp->input->nlcount = line - 1;
    cpp->line = line - 1;
    if ( file ) {
        cpp->input->file = file;
        cpp->file = file;
    }
}
