/******************************************************************************

    $Id: listfix.c,v 1.1 2003/08/28 09:53:28 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/listfix.c,v $

******************************************************************************/

#include "priv.h"


static PPTOKEN* cpp_paste_spaces(CPP* cpp, PPTOKEN* list)
{
    BUFFER*     buf = &cpp->scratch;
    const char* newline;
    const char* pos;
    size_t      last_nl = 0;

    cpp_buffer_clear(buf);
    while ( list ) {
        pos = list->token.str;
        while ( (newline = strchr(pos,'\n')) != NULL ) {
            buf->len = last_nl;
            cpp_buffer_catc(buf, '\n');
            last_nl++;
            pos = newline + 1;
        }
        cpp_buffer_cats(buf, pos);
        list = list->next;
    }

    return cpp_token_news(cpp, T_SPACE, cpp_buffer_gets(buf));
}


static PPTOKEN* cpp_paste_strings(CPP* cpp, PPTOKEN* list)
{
    BUFFER*     buf = &cpp->scratch;
    PPTOKEN*    token;
    const char* pos;
    bool        wide = false;
    bool        narrow = false;

    if ( list == NULL ) {
        return NULL;
    }
    if ( list->next == NULL ) {
        return list;
    }

    for ( token = list ; token ; token = token->next ) {
        if ( *token->token.str == 'L' ) {
            wide = true;
        } else {
            narrow = true;
        }
    }

    cpp_buffer_clear(buf);
    if ( wide ) {
        cpp_buffer_catc(buf, 'L');
    }
    cpp_buffer_catc(buf, '"');

    for ( token = list ; token ; token = token->next ) {
        pos = token->token.str;
        if ( *pos == 'L' ) {
            ++pos;
        }
        assert(*pos == '"');
        ++pos;
        while ( *pos && *pos != '"' ) {
            if ( pos[0] == '\\' && pos[1] ) {
                cpp_buffer_catc(buf, *pos++);
                if ( ISODIGIT(pos[0]) ) {
                    if ( pos[1] == '"' ) {
                        cpp_buffer_catc(buf, '0');
                        cpp_buffer_catc(buf, '0');
                    } else if ( ISODIGIT(pos[1]) && pos[2]=='"' ) {
                        cpp_buffer_catc(buf, '0');
                    }
                }
            }
            cpp_buffer_catc(buf, *pos++);
        }
    }

    cpp_buffer_catc(buf, '"');

    if ( wide && narrow ) {
        token = cpp_token_news(cpp, T_ERROR, "char/wchar_t mismatch in string paste");
        cpp_output_append(cpp, token);
    }

    return cpp_token_news(cpp, T_STRLIT, cpp_buffer_gets(buf));
}


static void cpp_paste_strings_and_spaces(CPP* cpp, PPLIST* list)
{
    PPTOKEN*    token;
    PPLIST      strings;
    PPLIST      spaces;
    PPLIST      other;
    PPLIST      result;

    cpp_list_reset(&result);
    while ( (token = cpp_list_getfirst(list)) != NULL ) {
        switch ( token->token.tag ) {
            case T_SPACE:
                cpp_list_reset(&spaces);
                cpp_list_reset(&other);
                cpp_list_append(&spaces, token);
                while ( list->first ) {
                    int tag = list->first->token.tag;
                    if ( tag == T_SPACE ) {
                        cpp_list_append(&spaces, cpp_list_getfirst(list));
                    } else if ( tag == T_PRAGMA || tag == T_ERROR ) {
                        cpp_list_append(&other, cpp_list_getfirst(list));
                    } else {
                        break;
                    }
                }
                token = cpp_paste_spaces(cpp, spaces.first);
                cpp_list_append(&result, token);
                cpp_list_appendlist(&result, other.first);
                break;
            case T_STRLIT:
                cpp_list_reset(&strings);
                cpp_list_reset(&spaces);
                cpp_list_reset(&other);
                cpp_list_append(&strings, token);
                while ( list->first ) {
                    int tag = list->first->token.tag;
                    if ( tag == T_STRLIT ) {
                        cpp_list_append(&strings, cpp_list_getfirst(list));
                    } else if ( tag == T_SPACE ) {
                        cpp_list_append(&spaces, cpp_list_getfirst(list));
                    } else if ( tag == T_PRAGMA || tag == T_ERROR ) {
                        cpp_list_append(&other, cpp_list_getfirst(list));
                    } else {
                        break;
                    }
                }
                token = cpp_paste_strings(cpp, strings.first);
                cpp_list_append(&result, token);
                cpp_list_appendlist(&result, other.first);
                if ( spaces.first ) {
                    token = cpp_paste_spaces(cpp, spaces.first);
                    cpp_list_append(&result, token);
                }
                break;
            default:
                cpp_list_append(&result, token);
                break;
        }
    }

    *list = result;
}


void cpp_listfix(CPP* cpp, PPLIST* list)
{
    cpp_paste_strings_and_spaces(cpp, list);
    cpp_extract_pragma_operator(cpp, list);
}
