/******************************************************************************

    $Id: paste.c,v 1.2 2003/09/05 20:30:27 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/paste.c,v $

******************************************************************************/

#include "priv.h"


PPTOKEN* cpp_paste_spaces(CPP* cpp, PPTOKEN* list)
{
    BUFFER*     buf = &cpp->scratch;
    const char* newline;
    const char* pos;
    size_t      last_nl = 0;

    cpp_buffer_clear(buf);
    while ( list ) {
        pos = list->token.str;
        while ( (newline = strchr(pos,'\n')) != NULL ) {
            buf->len = last_nl;
            cpp_buffer_catc(buf, '\n');
            last_nl++;
            pos = newline + 1;
        }
        cpp_buffer_cats(buf, pos);
        list = list->next;
    }

    return cpp_token_news(cpp, T_SPACE, cpp_buffer_gets(buf));
}


PPTOKEN* cpp_paste_strings(CPP* cpp, PPTOKEN* list)
{
    BUFFER*     buf = &cpp->scratch;
    PPTOKEN*    token;
    const char* pos;
    bool        wide = false;
    bool        narrow = false;

    for ( token = list ; token ; token = token->next ) {
        if ( *token->token.str == 'L' ) {
            wide = true;
        } else {
            narrow = true;
        }
    }

    cpp_buffer_clear(buf);
    if ( wide ) {
        cpp_buffer_catc(buf, 'L');
    }
    cpp_buffer_catc(buf, '"');

    for ( token = list ; token ; token = token->next ) {
        pos = token->token.str;
        if ( *pos == 'L' ) {
            ++pos;
        }
        assert(*pos == '"');
        ++pos;
        while ( *pos && *pos != '"' ) {
            if ( pos[0] == '\\' && pos[1] ) {
                cpp_buffer_catc(buf, *pos++);
                if ( ISODIGIT(pos[0]) ) {
                    if ( pos[1] == '"' ) {
                        cpp_buffer_catc(buf, '0');
                        cpp_buffer_catc(buf, '0');
                    } else if ( ISODIGIT(pos[1]) && pos[2]=='"' ) {
                        cpp_buffer_catc(buf, '0');
                    }
                }
            }
            cpp_buffer_catc(buf, *pos++);
        }
    }

    cpp_buffer_catc(buf, '"');

    if ( wide && narrow ) {
        token = cpp_token_news(cpp, T_ERROR, "char/wchar_t mismatch in string paste");
        cpp_output_append(cpp, token); // TODO: will this work when evaluating output list?
    }

    return cpp_token_news(cpp, T_STRLIT, cpp_buffer_gets(buf));
}


// TODO: handle _Pragma

void cpp_paste(CPP* cpp, PPLIST* list)
{
    PPTOKEN*    token;
    PPTOKEN*    tmp;
    PPLIST      strings;
    PPLIST      spaces;

    token = list->first;
    cpp_list_reset(list);
    while ( token ) {
        switch ( token->token.tag ) {
            case T_SPACE:
                cpp_list_reset(&spaces);
                while ( token && token->token.tag == T_SPACE ) {
                    tmp = token;
                    token = token->next;
                    cpp_list_append(&spaces, tmp);
                }
                tmp = cpp_paste_spaces(cpp, spaces.first);
                cpp_list_append(list, tmp);
                break;
            case T_STRLIT:
                cpp_list_reset(&strings);
                cpp_list_reset(&spaces);
                while ( token && (token->token.tag == T_STRLIT || token->token.tag == T_SPACE) ) {
                    tmp = token;
                    token = token->next;
                    if ( tmp->token.tag == T_STRLIT ) {
                        cpp_list_append(&strings, tmp);
                    } else {
                        cpp_list_append(&spaces, tmp);
                    }
                }
                tmp = cpp_paste_strings(cpp, strings.first);
                cpp_list_append(list, tmp);
                if ( spaces.first ) {
                    tmp = cpp_paste_spaces(cpp, spaces.first);
                    cpp_list_append(list, tmp);
                }
                break;
            default:
                tmp = token;
                token = token->next;
                cpp_list_append(list, tmp);
                break;
        }
    }
}

