/******************************************************************************

    $Id: pragma.c,v 1.4 2003/09/05 20:30:27 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/pragma.c,v $

******************************************************************************/

#include "priv.h"


static bool cpp_is_pragma(PPTOKEN* list)
{
    if ( list->token.tag != T_IDENT && list->token.tag != T_FINAL_IDENT ) {
        return false;
    }
    if ( strcmp(list->token.str, "_Pragma") ) {
        return false;
    }
    list = cpp_skip_t_space(list->next);
    if ( list == NULL || list->token.tag != T_PAREN_OPEN ) {
        return false;
    }
    list = cpp_skip_t_space(list->next);
    if ( list == NULL || list->token.tag != T_STRLIT ) {
        return false;
    }
    list = cpp_skip_t_space(list->next);
    if ( list == NULL || list->token.tag != T_PAREN_CLOSE ) {
        return false;
    }
    return true;
}


static void cpp_emit_pragma(CPP* cpp, PPTOKEN* strlit)
{
    PPTOKEN*    token;
    const char* pos;
    
    cpp_buffer_clear(&cpp->scratch);
    pos = strlit->token.str;
    if ( *pos == 'L' ) {
        ++pos;
    }
    assert(*pos == '"');
    ++pos;
    while ( *pos && *pos != '"' ) {
        if ( pos[0]=='\\' && (pos[1]=='"' || pos[1]=='\\') ) {
            ++pos;
        }
        cpp_buffer_catc(&cpp->scratch, *pos++);
    }

    token = cpp_token_news(cpp, T_PRAGMA, cpp_buffer_gets(&cpp->scratch));
    cpp_output_append(cpp, token);
}


void cpp_extract_pragma_operator(CPP* cpp, PPLIST* list)
{
    PPLIST      result;
    PPTOKEN*    token;

    cpp_list_reset(&result);

    while ( list->first ) {
        if ( !cpp_is_pragma(list->first) ) {
            cpp_list_append(&result, cpp_list_getfirst(list));
        } else {
            do {
                token = cpp_list_getfirst(list);
            } while ( token->token.tag != T_STRLIT);
            cpp_emit_pragma(cpp, token);
            do {
                token = cpp_list_getfirst(list);
            } while ( token->token.tag != T_PAREN_CLOSE );
        }
    }

    *list = result;
}


void cpp_parse_pragma(CPP* cpp, const char* pos)
{
    PPTOKEN*    list;
    PPTOKEN*    token;
    int         tag;
    size_t      len;

    pos = cpp_skipspace(pos);

    cpp_identify_token(pos, &tag, &len);
    if ( tag == T_END ) {
        return;
    }

    if ( (tag == T_IDENT && len==4 && memcmp(pos,"STDC",4)==0)
      || !cpp->expand_pragma
       ) 
    {
        // Do not perform macro expansion, but trim trailing spaces
        len = strlen(pos);
        while ( ISSPACE(pos[len-1]) ) {
            --len;
        }
        token = cpp_token_newn(cpp, T_PRAGMA, pos, len);
    } else {
        // Expand macros
        list = cpp_tokenize(cpp, pos);
        list = cpp_expand_list(cpp, list, NULL);
        cpp_buffer_clear(&cpp->scratch);
        cpp_list_flatten(&cpp->scratch, list);
        token = cpp_token_news(cpp, T_PRAGMA, cpp_buffer_gets(&cpp->scratch));
    }

    cpp_output_append(cpp, token);
}

