/******************************************************************************

    $Id: token.c,v 1.5 2003/08/28 21:59:29 mmp Exp $
    $Source: /archive/cvs/cc/src/cc/cpp/token.c,v $

******************************************************************************/

#include "priv.h"


/******************************************************************************
    Token creation
******************************************************************************/

PPTOKEN* cpp_token_new(CPP* cpp, int tag)
{
    PPTOKEN* token;

    token = cpp_heap_calloc(&cpp->output_heap, 1, sizeof(PPTOKEN));
    token->token.file = cpp->file;
    token->token.line = cpp->line;
    token->token.tag  = tag;
    token->token.str  = "";

    return token;
}


PPTOKEN* cpp_token_news(CPP* cpp, int tag, const char* str)
{
    PPTOKEN* token;
    
    token = cpp_token_new(cpp, tag);
    token->token.str = cpp_heap_strdup(&cpp->output_heap, str);

    return token;
}


PPTOKEN* cpp_token_news_recycle(CPP* cpp, int tag, const char* str)
{
    PPTOKEN* token;
    
    token = cpp_token_new(cpp, tag);
    token->token.str = str;

    return token;
}


PPTOKEN* cpp_token_newn(CPP* cpp, int tag, const char* str, size_t len)
{
    PPTOKEN* token;
    
    token = cpp_token_new(cpp, tag);
    token->token.str = cpp_heap_strndup(&cpp->output_heap, str, len);

    return token;
}


PPTOKEN* cpp_token_clone(CPP* cpp, PPTOKEN* token)
{
    PPTOKEN* clone;

    clone = cpp_heap_malloc(&cpp->output_heap, sizeof(PPTOKEN));
    memcpy(clone, token, sizeof(PPTOKEN));
    clone->next = NULL;

    return clone;
}


PPTOKEN* cpp_token_newnl(CPP* cpp, size_t count)
{
    char*       str;
    PPTOKEN*    token;
    
    str = cpp_heap_malloc(&cpp->output_heap, count+1);
    memset(str, '\n', count);
    str[count] = '\0';

    token = cpp_token_new(cpp, T_SPACE);
    token->token.str = str;

    return token;
}


PPTOKEN* cpp_tokenize(CPP* cpp, const char* pos)
{
    PPLIST  list;
    int     tag;
    size_t  len;
    bool    space = false;

    cpp_list_reset(&list);
    while ( cpp_identify_token(pos,&tag,&len), tag != T_END ) {
        if ( tag == T_SPACE ) {
            space = true;
        } else {
            if ( space ) {
                cpp_list_append(&list, cpp_token_news_recycle(cpp, T_SPACE, " "));
                space = false;
            }
            cpp_list_append(&list, cpp_token_newn(cpp, tag, pos, len));
        }
        pos += len;
    }

    return list.first;
}


/******************************************************************************
    List handling
******************************************************************************/

PPTOKEN* cpp_skip_t_space(PPTOKEN* list)
{
    while ( list && list->token.tag == T_SPACE ) {
        list = list->next;
    }
    return list;
}


PPTOKEN* cpp_reverse_tokens(PPTOKEN* tokens)
{
    PPTOKEN* token;
    PPTOKEN* list = NULL;

    while ( tokens ) {
        token = tokens;
        tokens = token->next;
        token->next = list;
        list = token;
    }

    return list;
}


void cpp_list_reset(PPLIST* list)
{
    memset(list, 0, sizeof(PPLIST));
}


void cpp_list_append(PPLIST* list, PPTOKEN* token)
{
    token->next = NULL;
    if ( list->first ) {
        list->last->next = token;
    } else {
        list->first = token;
    }
    list->last = token;
}


void cpp_list_appendlist(PPLIST* list, PPTOKEN* tokens)
{
    PPTOKEN* token;

    while ( (token=tokens) != NULL ) {
        tokens = token->next;
        cpp_list_append(list, token);
    }
}


PPTOKEN* cpp_list_getfirst(PPLIST* list)
{
    PPTOKEN* token = list->first;

    if ( token ) {
        list->first = token->next;
        if ( list->first == NULL ) {
            list->last = NULL;
        }
    }

    return token;
}
